/****************************************************************************
 * Functions for dealing with window borders.
 * Copyright (C) 2004 Joe Wingbermuehle
 ****************************************************************************/

#include "jwm.h"

static char x_bitmap[] = {
	0x00, 0x00, 0x00, 0x00, 0x0C, 0x30, 0x18, 0x18, 0x30, 0x0C, 0x60, 0x06,
	0xC0, 0x03, 0x80, 0x01, 0xC0, 0x03, 0x60, 0x06, 0x30, 0x0C, 0x18, 0x18,
	0x0C, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static char min_bitmap[] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x3F,
	0xFC, 0x3F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static char max_bitmap[] = {
	0x00, 0x00, 0x00, 0x00, 0xFC, 0x3F, 0xFC, 0x3F, 0x04, 0x20, 0x04, 0x20,
	0x04, 0x20, 0x04, 0x20, 0x04, 0x20, 0x04, 0x20, 0x04, 0x20, 0x04, 0x20,
	0xFC, 0x3F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static char maxa_bitmap[] = {
	0x00, 0x00, 0x00, 0x00, 0xFC, 0x3F, 0xFC, 0x3F, 0x04, 0x20, 0xE4, 0x27,
	0xE4, 0x27, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0xE4, 0x27, 0x04, 0x20,
	0xFC, 0x3F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static void DrawBorderHelper(ClientNode *np, unsigned int height);
static void DrawButtonBorder(Drawable canvas, GC gc, ClientNode *np,
	int offset);

/****************************************************************************
 ****************************************************************************/
BorderActionType GetBorderActionType(ClientNode *np, int x, int y) {
	int north;
	int offset;

	if(!(np->borderFlags & BORDER_OUTLINE)) {
		return BA_NONE;
	}

	if(np->borderFlags & BORDER_TITLE) {

		if(y > borderWidth && y <= borderWidth + EDGE_SIZE) {
			offset = np->width + borderWidth - EDGE_SIZE;
			if(np->borderFlags & BORDER_CLOSE) {
				if(x > offset && x < offset + EDGE_SIZE) {
					return BA_CLOSE;
				}
				offset -= EDGE_SIZE;
			}
			if(np->borderFlags & BORDER_MAX) {
				if(x > offset && x < offset + EDGE_SIZE) {
					return BA_MAXIMIZE;
				}
				offset -= EDGE_SIZE;
			}
			if(np->borderFlags & BORDER_MIN) {
				if(x > offset && x < offset + EDGE_SIZE) {
					return BA_MINIMIZE;
				}
			}
		}

		if(y >= borderWidth && y <= titleSize) {
			if(x >= borderWidth && x < np->width - borderWidth) {
				if(np->borderFlags & BORDER_MOVE) {
					return BA_MOVE;
				} else {
					return BA_NONE;
				}
			}
		}

		north = titleSize;
	} else {
		north = borderWidth;
	}

	if(!(np->borderFlags & BORDER_RESIZE)) {
		return BA_NONE;
	}

	if(x < borderWidth + EDGE_SIZE && y < EDGE_SIZE + borderWidth) {
		return  BA_RESIZE_N | BA_RESIZE_W | BA_RESIZE;
	} else if(x < EDGE_SIZE + borderWidth
		&& y - north >= np->height - EDGE_SIZE) {
		return BA_RESIZE_S | BA_RESIZE_W | BA_RESIZE;
	} else if(x - borderWidth >= np->width - EDGE_SIZE
		&& y < EDGE_SIZE + borderWidth) {
		return BA_RESIZE_N | BA_RESIZE_E | BA_RESIZE;
	} else if(x - borderWidth >= np->width - EDGE_SIZE
		&& y - north >= np->height - EDGE_SIZE) {
		return BA_RESIZE_S | BA_RESIZE_E | BA_RESIZE;
	} else if(x < borderWidth) {
		return BA_RESIZE_W | BA_RESIZE;
	} else if(x >= np->width + borderWidth) {
		return BA_RESIZE_E | BA_RESIZE;
	} else if(y < borderWidth) {
		return BA_RESIZE_N | BA_RESIZE;
	} else if(y >= np->height) {
		return BA_RESIZE_S | BA_RESIZE;
	} else {
		return BA_NONE;
	}
}

/****************************************************************************
 ****************************************************************************/
void DrawBorder(ClientNode *np) {
	unsigned int height;

	if(shouldExit) {
		return;
	}

	if(np->statusFlags & (STAT_MINIMIZED | STAT_WITHDRAWN | STAT_HIDDEN)) {
		return;
	}

	if(!(np->borderFlags & BORDER_TITLE)
		&& !(np->borderFlags & BORDER_OUTLINE)) {
		return;
	}

	if(np->statusFlags & STAT_SHADED) {
		height = titleSize + borderWidth;
	} else if(np->borderFlags & BORDER_TITLE) {
		height = np->height + titleSize + borderWidth;
	} else {
		height = np->height + 2 * borderWidth;
	}

	DrawBorderHelper(np, height);

}

/****************************************************************************
 ****************************************************************************/
void DrawBorderHelper(ClientNode *np, unsigned int height) {
	unsigned int width;
	long borderPixel, borderTextPixel;
	long pixelUp, pixelDown;
	int buttonCount, titleWidth;
	XRectangle rect;
	Pixmap canvas;
	GC canvasGC;

	width = np->width + borderWidth + borderWidth;

	if(np->statusFlags & STAT_ACTIVE) {
		borderPixel = colors[COLOR_ACTIVE_BORDER];
		borderTextPixel = colors[COLOR_ACTIVE_BORDER_TEXT];
		pixelUp = colors[COLOR_ACTIVE_BORDER_UP];
		pixelDown = colors[COLOR_ACTIVE_BORDER_DOWN];
	} else {
		borderPixel = colors[COLOR_BORDER];
		borderTextPixel = colors[COLOR_BORDER_TEXT];
		pixelUp = colors[COLOR_BORDER_UP];
		pixelDown = colors[COLOR_BORDER_DOWN];
	}

	canvas = XCreatePixmap(display, np->parent, width, height,
		DefaultDepth(display, DefaultScreen(display)));
	canvasGC = XCreateGC(display, canvas, 0, NULL);

	XSetForeground(display, canvasGC, borderPixel);
	XFillRectangle(display, canvas, canvasGC, 2, 2,
		width - 4, height - 4);

	buttonCount = DrawBorderButtons(np, canvas, canvasGC);
	titleWidth = width - EDGE_SIZE * buttonCount - borderWidth * 2 - 2;

	if(np->name && np->name[0] && (np->borderFlags & BORDER_TITLE)) {

		rect.x = borderWidth + 1;
		rect.y = borderWidth + 1;
		if(titleWidth > 0) {
			rect.width = titleWidth;
		} else {
			rect.width = 0;
		}
		rect.height = EDGE_SIZE;
		XSetClipRectangles(display, canvasGC, 0, 0, &rect, 1, Unsorted);

		XSetForeground(display, canvasGC, borderTextPixel);
		XSetBackground(display, canvasGC, borderPixel);
		XSetFont(display, canvasGC, borderFont->fid);
		XDrawString(display, canvas, canvasGC,
			borderWidth + 4,
			borderFont->ascent + borderFont->descent + borderWidth - 2,
			np->name, strlen(np->name));

		XSetClipMask(display, canvasGC, None);

	}

	/* Draw title outline */
	XSetForeground(display, canvasGC, pixelUp);
	XDrawLine(display, canvas, canvasGC, borderWidth, borderWidth,
		width - borderWidth - 1, borderWidth);
	XDrawLine(display, canvas, canvasGC, borderWidth, borderWidth + 1,
		borderWidth, EDGE_SIZE + borderWidth - 1);

	XSetForeground(display, canvasGC, pixelDown);
	XDrawLine(display, canvas, canvasGC, borderWidth + 1,
		EDGE_SIZE + borderWidth - 1, width - borderWidth + 2,
		EDGE_SIZE + borderWidth - 1);
	XDrawLine(display, canvas, canvasGC, width - borderWidth - 1,
		borderWidth + 1, width - borderWidth - 1, EDGE_SIZE + borderWidth);

	/* Draw outline */
	XSetForeground(display, canvasGC, pixelUp);
	XDrawLine(display, canvas, canvasGC, width - borderWidth, borderWidth,
		width - borderWidth, height - borderWidth);
	XDrawLine(display, canvas, canvasGC, borderWidth, height - borderWidth,
		width - borderWidth, height - borderWidth);

	XSetForeground(display, canvasGC, pixelDown);
	XDrawLine(display, canvas, canvasGC, borderWidth - 1, borderWidth - 1,
		width - borderWidth, borderWidth - 1);
	XDrawLine(display, canvas, canvasGC, borderWidth - 1, borderWidth,
		borderWidth - 1, height - borderWidth);

	XFillRectangle(display, canvas, canvasGC, width - 2, 0, 2, height);
	XFillRectangle(display, canvas, canvasGC, 0, height - 2, width, 2);
	XSetForeground(display, canvasGC, pixelUp);
	XDrawLine(display, canvas, canvasGC, 0, 0, 0, height - 1);
	XDrawLine(display, canvas, canvasGC, 1, 1, 1, height - 2);
	XDrawLine(display, canvas, canvasGC, 1, 0, width - 1, 0);
	XDrawLine(display, canvas, canvasGC, 1, 1, width - 2, 1);

	/* Draw marks */
	XSetForeground(display, canvasGC, pixelDown);

	/* Upper left */
	XDrawLine(display, canvas, canvasGC, EDGE_SIZE + borderWidth - 1,
		2, EDGE_SIZE + borderWidth - 1, borderWidth - 2);
	XDrawLine(display, canvas, canvasGC, 2, EDGE_SIZE + borderWidth - 1,
		borderWidth - 2, EDGE_SIZE + borderWidth - 1);

	/* Upper right */
	XDrawLine(display, canvas, canvasGC, width - EDGE_SIZE - borderWidth - 1,
		2, width - EDGE_SIZE - borderWidth - 1, borderWidth - 2);
	XDrawLine(display, canvas, canvasGC, width - 3,
		EDGE_SIZE + borderWidth - 1, width - borderWidth + 1,
		EDGE_SIZE + borderWidth - 1);

	/* Lower left */
	XDrawLine(display, canvas, canvasGC, 2,
		height - EDGE_SIZE - borderWidth - 1, borderWidth - 2,
		height - EDGE_SIZE - borderWidth - 1);
	XDrawLine(display, canvas, canvasGC, EDGE_SIZE + borderWidth - 1,
		height - 3, EDGE_SIZE + borderWidth - 1, height - borderWidth + 1);

	/* Lower right */
	XDrawLine(display, canvas, canvasGC, width - 3,
		height - EDGE_SIZE - borderWidth - 1, width - borderWidth + 1,
		height - EDGE_SIZE - borderWidth - 1);
	XDrawLine(display, canvas, canvasGC, width - EDGE_SIZE - borderWidth - 1,
		height - 3, width - EDGE_SIZE - borderWidth - 1,
		height - borderWidth + 1);

	XSetForeground(display, canvasGC, pixelUp);

	/* Upper left */
	XDrawLine(display, canvas, canvasGC, EDGE_SIZE + borderWidth,
		2, EDGE_SIZE + borderWidth, borderWidth - 2);
	XDrawLine(display, canvas, canvasGC, 2, EDGE_SIZE + borderWidth,
		borderWidth - 2, EDGE_SIZE + borderWidth);

	/* Upper right */
	XDrawLine(display, canvas, canvasGC, width - EDGE_SIZE - borderWidth, 2,
		width - EDGE_SIZE - borderWidth, borderWidth - 2);
	XDrawLine(display, canvas, canvasGC, width - 3, EDGE_SIZE + borderWidth,
		width - borderWidth + 1, EDGE_SIZE + borderWidth);

	/* Lower left */
	XDrawLine(display, canvas, canvasGC, 2, height - EDGE_SIZE - borderWidth,
		borderWidth - 2, height - EDGE_SIZE - borderWidth);
	XDrawLine(display, canvas, canvasGC, EDGE_SIZE + borderWidth,
		height - 3, EDGE_SIZE + borderWidth, height - borderWidth + 1);

	/* Lower right */
	XDrawLine(display, canvas, canvasGC, width - 3,
		height - EDGE_SIZE - borderWidth, width - borderWidth + 1,
		height - EDGE_SIZE - borderWidth);
	XDrawLine(display, canvas, canvasGC, width - EDGE_SIZE - borderWidth,
		height - 3, width - EDGE_SIZE - borderWidth,
		height - borderWidth + 1);

	XCopyArea(display, canvas, np->parent, np->parentGC, 0, 0,
		width, height, 0, 0);

	XFreeGC(display, canvasGC);
	XFreePixmap(display, canvas);

}

/****************************************************************************
 ****************************************************************************/
void DrawButtonBorder(Drawable canvas, GC gc, ClientNode *np, int offset) {
	long up, down;

	if(np->statusFlags & STAT_ACTIVE) {
		up = colors[COLOR_ACTIVE_BORDER_UP];
		down = colors[COLOR_ACTIVE_BORDER_DOWN];
	} else {
		up = colors[COLOR_BORDER_UP];
		down = colors[COLOR_BORDER_DOWN];
	}

	XSetForeground(display, gc, up);
	XDrawLine(display, canvas, gc, offset, borderWidth + 1, offset,
		EDGE_SIZE + borderWidth  - 2);

	XSetForeground(display, gc, down);
	XDrawLine(display, canvas, gc, offset - 1, borderWidth + 1, offset - 1,
		EDGE_SIZE + borderWidth - 2);

}

/****************************************************************************
 ****************************************************************************/
int DrawBorderButtons(ClientNode *np, Drawable canvas, GC gc) {
	Pixmap pixmap;
	int count = 0;
	unsigned int offset;
	long fg, bg;

	if(!(np->borderFlags & BORDER_TITLE)) {
		return count;
	}

	offset = np->width + borderWidth - EDGE_SIZE;

	if(np->statusFlags & STAT_ACTIVE) {
		fg = colors[COLOR_ACTIVE_BORDER_TEXT];
		bg = colors[COLOR_ACTIVE_BORDER];
	} else {
		fg = colors[COLOR_BORDER_TEXT];
		bg = colors[COLOR_BORDER];
	}

	if(np->borderFlags & BORDER_CLOSE) {
		DrawButtonBorder(canvas, gc, np, offset);

		pixmap = XCreatePixmapFromBitmapData(display, canvas, x_bitmap,
			16, 16, fg, bg, DefaultDepth(display, DefaultScreen(display)));
		XCopyArea(display, pixmap, canvas, gc, 0, 0, 16, 16,
			offset + 2, borderWidth + 3);
		XFreePixmap(display, pixmap);

		offset -= EDGE_SIZE;
		++count;
	}

	if(np->borderFlags & BORDER_MAX) {

		if(np->statusFlags & STAT_MAXIMIZED) {
			pixmap = XCreatePixmapFromBitmapData(display, canvas, maxa_bitmap,
				16, 16, fg, bg, DefaultDepth(display, DefaultScreen(display)));
		} else {
			pixmap = XCreatePixmapFromBitmapData(display, canvas, max_bitmap,
				16, 16, fg, bg, DefaultDepth(display, DefaultScreen(display)));
		}
		XCopyArea(display, pixmap, canvas, gc, 0, 0, 16, 16,
			offset + 2, borderWidth + 3);
		XFreePixmap(display, pixmap);

		DrawButtonBorder(canvas, gc, np, offset);

		offset -= EDGE_SIZE;
		++count;
	}

	if(np->borderFlags & BORDER_MIN) {
		DrawButtonBorder(canvas, gc, np, offset);
		pixmap = XCreatePixmapFromBitmapData(display, canvas, min_bitmap,
			16, 16, fg, bg, DefaultDepth(display, DefaultScreen(display)));
		XCopyArea(display, pixmap, canvas, gc, 0, 0, 16, 16,
			offset + 2, borderWidth + 3);
		XFreePixmap(display, pixmap);
		++count;
	}

	return count;

}

/****************************************************************************
 ****************************************************************************/
void SetBorderWidth(const char *str) {
	int width;

	width = atoi(str);
	if(width < MIN_BORDER_WIDTH || width > MAX_BORDER_WIDTH) {
		borderWidth = DEFAULT_BORDER_WIDTH;
		Warning("invalid border width specified: %d", width);
	} else {
		borderWidth = width;
	}
	titleSize = DEFAULT_TITLE_SIZE - DEFAULT_BORDER_WIDTH + borderWidth;

}

