/****************************************************************************
 * Functions to handle XServer events.
 * Copyright (C) 2004 Joe Wingbermuehle
 ****************************************************************************/

#include "jwm.h"

static void DispatchBorderButtonEvent(XButtonEvent *event, ClientNode *np);

static void HandleConfigureRequest(XConfigureRequestEvent *event);
static void HandleExpose(XExposeEvent *event);
static void HandlePropertyNotify(XPropertyEvent *event);
static void HandleClientMessage(XClientMessageEvent *event);
static void HandleShapeEvent(XShapeEvent *event);
static void HandleColormapChange(XColormapEvent *event);
static void HandleDestroyNotify(XDestroyWindowEvent *event);
static void HandleMapRequest(XMapEvent *event);
static void HandleUnmapNotify(XUnmapEvent *event);

/****************************************************************************
 ****************************************************************************/
void WaitForEvent(XEvent *event) {
	struct timeval timeout;
	int fd;
	int handled;
	fd_set fds;

	fd = XConnectionNumber(display);

	do {

		while(XPending(display) == 0) {
			FD_ZERO(&fds);
			FD_SET(fd, &fds);
			timeout.tv_sec = MAIN_DELTA;
			timeout.tv_usec = 0;
			if(select(fd + 1, &fds, NULL, NULL, &timeout) <= 0) {
				UpdateTime();
			}
		}

		XNextEvent(display, event);

		switch(event->type) {
		case ConfigureRequest:
			HandleConfigureRequest(&event->xconfigurerequest);
			handled = 1;
			break;
		case MapRequest:
			HandleMapRequest(&event->xmap);
			handled = 1;
			break;
		case PropertyNotify:
			HandlePropertyNotify(&event->xproperty);
			handled = 1;
			break;
		case ClientMessage:
			HandleClientMessage(&event->xclient);
			handled = 1;
			break;
		case UnmapNotify:
			HandleUnmapNotify(&event->xunmap);
			handled = 1;
			break;
		case Expose:
			HandleExpose(&event->xexpose);
			handled = 1;
			break;
		case ColormapNotify:
			HandleColormapChange(&event->xcolormap);
			handled = 1;
			break;
		case DestroyNotify:
			HandleDestroyNotify(&event->xdestroywindow);
			handled = 1;
			break;
		case ConfigureNotify:
		case CreateNotify:
		case MapNotify:
		case NoExpose:
		case ReparentNotify:
			handled = 1;
			break;
		default:
#ifdef USE_SHAPE
			if(haveShape && event->type == shapeEvent) {
				HandleShapeEvent((XShapeEvent*)&event);
				handled = 1;
			} else {
				handled = 0;
			}
#else
			handled = 0;
#endif
			break;
		}

	} while(handled);

}

/****************************************************************************
 ****************************************************************************/
void HandleButtonPress(XButtonEvent *event) {
	int x, y;
	ClientNode *np;

	np = FindClientByParent(event->window);
	if(np) {
		RaiseClient(np);
		switch(event->button) {
		case Button1:
			DispatchBorderButtonEvent(event, np);
			break;
		case Button2:
			MoveClient(np);
			break;
		case Button3:
			x = event->x + np->x;
			y = event->y + np->y;
			if(np->borderFlags & BORDER_OUTLINE) {
				x -= borderWidth;
				if(np->borderFlags & BORDER_TITLE) {
					y -= titleSize;
				} else {
					y -= borderWidth;
				}
			} else {
				if(np->borderFlags & BORDER_TITLE) {
					y -= titleSize;
				}
			}
			ShowWindowMenu(np, x, y);
			break;
		default:
			break;
		}
	} else if(event->window == trayWindow) {
		HandleTrayClick(event);
	} else if(event->window == rootWindow) {
		ShowRootMenu(event->x, event->y);
	} else {
		np = FindClientByWindow(event->window);
		if(np) {
			switch(event->button) {
			case Button1:
			case Button2:
			case Button3:
				RaiseClient(np);
				break;
			default:
				break;
			}
			XAllowEvents(display, ReplayPointer, CurrentTime);
		}
	}

	UpdatePager();
}

/****************************************************************************
 ****************************************************************************/
void HandleKeyPress(XKeyEvent *event) {
	ClientNode *np;
	KeyType key;

	key = GetKey(event);

	np = FindClientByWindow(event->window);

	switch(key) {
	case KEY_DESKTOP:
		ChangeDesktop((currentDesktop + 1) % desktopCount);
		break;
	case KEY_NEXT:
		FocusNext();
		break;
	case KEY_CLOSE:
		if(np) {
			DeleteClient(np);
		}
		break;
	case KEY_SHADE:
		if(np) {
			if(np->statusFlags & STAT_SHADED) {
				UnshadeClient(np);
			} else {
				ShadeClient(np);
			}
		}
		break;
	case KEY_MOVE:
		if(np) {
			MoveClientKeyboard(np);
		}
		break;
	case KEY_RESIZE:
		if(np) {
			ResizeClientKeyboard(np);
		}
		break;
	case KEY_MIN:
		if(np) {
			MinimizeClient(np);
		}
		break;
	case KEY_MAX:
		if(np) {
			MaximizeClient(np);
		}
		break;
	case KEY_ROOT:
		ShowRootMenu(0, 0);
		break;
	case KEY_WIN:
		if(np) {
			ShowWindowMenu(np, np->x, np->y);
		}
		break;
	default:
		break;
	}

}

/****************************************************************************
 ****************************************************************************/
void HandleConfigureRequest(XConfigureRequestEvent *event) {
	XWindowChanges wc;
	ClientNode *np;
	int north, south, east, west;
	int changed;

	np = FindClientByWindow(event->window);
	if(np && np->window == event->window) {

		if(np->controller) {
			(np->controller)(0);
		}

		changed = 0;
		if((event->value_mask & CWWidth) && (event->width != np->width)) {
			np->width = event->width;
			changed = 1;
		}
		if((event->value_mask & CWHeight) && (event->height != np->height)) {
			np->height = event->height;
			changed = 1;
		}
		if((event->value_mask & CWX) && (event->x != np->x)) {
			np->x = event->x;
			changed = 1;
		}
		if((event->value_mask & CWY) && (event->y != np->y)) {
			np->y = event->y;
			changed = 1;
		}

		if(!changed) {
			return;
		}

		if(np->borderFlags & BORDER_OUTLINE) {
			east = borderWidth;
			west = borderWidth;
			south = borderWidth;
			if(np->borderFlags & BORDER_TITLE) {
				north = titleSize;
			} else {
				north = borderWidth;
			}
		} else {
			north = 0;
			south = 0;
			east = 0;
			west = 0;
		}

		wc.stack_mode = Above;
		wc.sibling = np->parent;
		wc.border_width = 0;

		wc.x = west;
		wc.y = north;
		wc.width = np->width;
		wc.height = np->height;
		XConfigureWindow(display, np->window, event->value_mask, &wc);

		wc.x = np->x;
		wc.y = np->y;
		wc.width = np->width + east + west;
		wc.height = np->height + north + south;
		XConfigureWindow(display, np->parent, event->value_mask, &wc);

#ifdef USE_SHAPE
		if(np->statusFlags & STAT_USESHAPE) {
			SetShape(np);
		}
#endif

	} else {
		wc.stack_mode = event->detail;
		wc.sibling = event->above;
		wc.border_width = event->border_width;
		wc.x = event->x;
		wc.y = event->y;
		wc.width = event->width;
		wc.height = event->height;
		XConfigureWindow(display, event->window, event->value_mask, &wc);
	}

}

/****************************************************************************
 ****************************************************************************/
void HandleEnterNotify(XCrossingEvent *event) {
	ClientNode *np;
	Cursor cur;
	BorderActionType action;

	np = FindClientByWindow(event->window);
	if(np) {
		if(!(np->statusFlags & STAT_ACTIVE)) {
			FocusClient(np);
		}
		if(np->parent == event->window) {
			action = GetBorderActionType(np, event->x, event->y);
			np->borderAction = action;
			cur = GetFrameCursor(action);
			XDefineCursor(display, np->parent, cur);
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleLeaveNotify(XCrossingEvent *event) {
	ClientNode *np;

	np = FindClientByParent(event->window);
	if(np) {
		SetDefaultCursor(np->parent);
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleExpose(XExposeEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np) {
		DrawBorder(np);
	} else if(event->window == trayWindow) {
		DrawTray();
	}
}

/****************************************************************************
 ****************************************************************************/
void HandlePropertyNotify(XPropertyEvent *event) {
	ClientNode *np;
	char *atomName;

	np = FindClientByWindow(event->window);
	if(np) {
		switch(event->atom) {
		case XA_WM_NAME:
			ReadWMName(np);
			break;
		case XA_WM_NORMAL_HINTS:
			ReadWMNormalHints(np);
			break;
		case XA_WM_HINTS:
			break;
		case XA_WM_ICON_NAME:
		case XA_WM_CLIENT_MACHINE:
			break;
		default:
			if(event->atom == protWMState) {
			} else if(event->atom == protWMLocaleName) {
			} else if(event->atom == protNetWMName) {
			} else if(event->atom == protNetWMIcon) {
			} else if(event->atom == protWMCommand) {
			} else if(event->atom == protNetWMDesktop) {
				ReadNetWMDesktop(np);
			} else if(event->atom == protWinLayer) {
				ReadWinLayer(np);
			} else {
				atomName = XGetAtomName(display, event->atom);
				Debug("Uknown PropertyNotify: %s", atomName);
				XFree(atomName);
			}
			break;
		}

		DrawBorder(np);
		DrawTray();
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleClientMessage(XClientMessageEvent *event) {
	ClientNode *np;
	long mask, flags;
	char *atomName;

	np = FindClientByWindow(event->window);
	if(np) {
		if(event->message_type == protWinState) {

			mask = event->data.l[0];
			flags = event->data.l[1];

			if(mask & WIN_STATE_STICKY) {
				if(flags & WIN_STATE_STICKY) {
					SetClientSticky(np, 1);
				} else {
					SetClientSticky(np, 0);
				}
			}

			if(mask & WIN_STATE_HIDDEN) {
				if(flags & WIN_STATE_HIDDEN) {
					np->statusFlags |= STAT_NOLIST;
				} else {
					np->statusFlags &= ~STAT_NOLIST;
				}
				DrawTray();
			}

		} else if(event->message_type == protWinLayer) {

			np->layer = event->data.l[0];
			SetClientLayer(np, event->data.l[0]);

		} else if(event->message_type == protWMChangeState) {

			if(np->controller) {
				(np->controller)(0);
			}

			switch(event->data.l[0]) {
			case WithdrawnState:
				SetClientWithdrawn(np, 1);
				break;
			case IconicState:
				MinimizeClient(np);
				break;
			case NormalState:
				RestoreClient(np);
				break;
			default:
				break;
			}

		} else if(event->message_type == protNetActiveWindow) {

			if(event->data.l[0] == 0) {
				RestoreClient(np);
			}

		} else if(event->message_type == protNetWMDesktop) {

			if(event->data.l[0] == 0xFFFFFFFF) {
				SetClientSticky(np, 1);
			} else {

				if(np->controller) {
					(np->controller)(0);
				}

				if(event->data.l[0] >= 0 && event->data.l[0] < desktopCount) {
					np->statusFlags &= ~STAT_STICKY;
					SetClientDesktop(np, event->data.l[0]);
				}
			}

		} else {
			atomName = XGetAtomName(display, event->message_type);
			Debug("Uknown ClientMessage: %s", atomName);
			XFree(atomName);
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleMotionNotify(XMotionEvent *event) {
	ClientNode *np;
	Cursor cur;
	BorderActionType action;

	np = FindClientByParent(event->window);
	if(np && (np->borderFlags & BORDER_OUTLINE)) {
		action = GetBorderActionType(np, event->x, event->y);
		if(np->borderAction != action) {
			np->borderAction = action;
			cur = GetFrameCursor(action);
			XDefineCursor(display, np->parent, cur);
		}
	}

}
/****************************************************************************
 ****************************************************************************/
void HandleShapeEvent(XShapeEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np) {
		SetShape(np);
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleColormapChange(XColormapEvent *event) {
	ClientNode *np;

	if(event->new) {
		np = FindClientByWindow(event->window);
		if(np) {
			if(np->colorMap && np->statusFlags & STAT_ACTIVE) {
				XUninstallColormap(display, np->colorMap);
			}
			np->colorMap = event->colormap;
			if(np->colorMap && np->statusFlags & STAT_ACTIVE) {
				XInstallColormap(display, np->colorMap);
			}
		}
	}
}

/****************************************************************************
 ****************************************************************************/
void HandleMapRequest(XMapEvent *event) {
	ClientNode *np;

	Assert(event);

	np = FindClientByWindow(event->window);
	if(!np) {
		AddClientWindow(event->window, 0);
	} else {
		if(!(np->statusFlags & STAT_MAPPED)) {
			np->statusFlags |= STAT_MAPPED;
			np->statusFlags &= ~STAT_MINIMIZED;
			XMapWindow(display, np->window);
			XMapWindow(display, np->parent);
			RaiseClient(np);
			DrawTray();
		}
	}
	RestackClients();
}

/****************************************************************************
 ****************************************************************************/
void HandleUnmapNotify(XUnmapEvent *event) {
	ClientNode *np;

	Assert(event);

	np = FindClientByWindow(event->window);
	if(np && np->window == event->window) {

		if(np->controller) {
			(np->controller)(1);
		}

		if(np->statusFlags & STAT_MAPPED) {
			np->statusFlags &= ~STAT_MAPPED;
			XUnmapWindow(display, np->parent);
		}
	}

}

/****************************************************************************
 ****************************************************************************/
void HandleDestroyNotify(XDestroyWindowEvent *event) {
	ClientNode *np;

	np = FindClientByWindow(event->window);
	if(np && np->window == event->window) {

		if(np->controller) {
			(np->controller)(1);
		}

		RemoveClient(np);
	}

}

/****************************************************************************
 ****************************************************************************/
void DispatchBorderButtonEvent(XButtonEvent *event, ClientNode *np) {
	static Time lastClickTime = 0;
	static int lastX = 0, lastY = 0;
	static int doubleClickActive = 0;
	BorderActionType action;

	if(!(np->borderFlags & BORDER_OUTLINE)) {
		return;
	}

	action = GetBorderActionType(np, event->x, event->y);

	switch(action & 0x0F) {
	case BA_RESIZE:
		ResizeClient(np, action);
		break;
	case BA_MOVE:
		if(doubleClickActive
			&& Abs(event->time - lastClickTime) <= doubleClickSpeed
			&& Abs(event->x - lastX) <= doubleClickDelta
			&& Abs(event->y - lastY) <= doubleClickDelta) {
			if(np->statusFlags & STAT_SHADED) {
				UnshadeClient(np);
			} else {
				ShadeClient(np);
			}
			doubleClickActive = 0;
		} else {
			if(MoveClient(np)) {
				doubleClickActive = 0;
			} else {
				doubleClickActive = 1;
				lastClickTime = event->time;
				lastX = event->x;
				lastY = event->y;
			}
		}
		break;
	case BA_CLOSE:
		DeleteClient(np);
		break;
	case BA_MAXIMIZE:
		MaximizeClient(np);
		break;
	case BA_MINIMIZE:
		MinimizeClient(np);
		break;
	default:
		break;
	}

}

